import numpy as np


def im3DNORM(img, normind, varargin=None):
    """
    % IMAGE3DNORM computes the desired image norm
    %   IMAGE3DNORM(IMG,NORMIND) computes the norm of image IMG using the norm
    %   defined in NORMIND
    %
    %   IMG         A 3D image
    %   NORMIND     {non-zero int, inf, -inf, 'fro', 'nuc'}
    %               'TV': TV norm
    %
    %
    %--------------------------------------------------------------------------
    %--------------------------------------------------------------------------
    % This file is part of the TIGRE Toolbox
    %
    % Copyright (c) 2015, University of Bath and
    %                     CERN-European Organization for Nuclear Research
    %                     All rights reserved.
    %
    % License:            Open Source under BSD.
    %                     See the full license at
    %                     https://github.com/CERN/TIGRE/license.txt
    %
    % Contact:            tigre.toolbox@gmail.com
    % Codes:              https://github.com/CERN/TIGRE/
    % Coded by:           Ander Biguri
    %--------------------------------------------------------------------------
    """
    if normind is [np.inf, -np.inf, "fro", "nuc"]:
        return np.linalg.norm(img.ravel(), normind)
    if type(normind) is int:
        return np.linalg.norm(img.ravel(), normind)
    if normind == "TV":
        gx, gy, gz = np.diff(img, axis=2), np.diff(img, axis=1), np.diff(img, axis=0)
        g = np.sum(np.sqrt(gx * gx + gy * gy + gz * gz))
        return g



import copy

import numpy as np
from _Ax import _Ax_ext

from .gpu import GpuIds


def Ax(img, geo, angles, projection_type="Siddon", **kwargs):
    if img.dtype != np.float32:
        raise TypeError("Input data should be float32, not " + str(img.dtype))
    if not np.isreal(img).all():
        raise ValueError("Complex types not compatible for projection.")
    if any(img.shape != geo.nVoxel):
        raise ValueError(
            "Input data should be of shape geo.nVoxel: "
            + str(geo.nVoxel)
            + " not:"
            + str(img.shape)
        )
    geox = copy.deepcopy(geo)
    geox.check_geo(angles)
    """
    Here we cast all values in geo to single point precision float. This way we know what behaviour
    to expect from pytigre to Cuda and can change single parameters accordingly.
    """
    geox.cast_to_single()
    # geox.checknans()

    if kwargs.get("gpuids", None) is None:
        gpuids = GpuIds()
    else:
        gpuids = kwargs["gpuids"]

    return _Ax_ext(img, geox, geox.angles, projection_type, geox.mode, gpuids=gpuids)



from libc.stdlib cimport malloc, free
cdef extern from "types_TIGRE.hpp":
    ctypedef struct Geometry:
        # Parameters part of the image geometry
        int   nVoxelX, nVoxelY, nVoxelZ;
        float sVoxelX, sVoxelY, sVoxelZ;
        float dVoxelX, dVoxelY, dVoxelZ;
        float *offOrigX;
        float *offOrigY;
        float *offOrigZ;
        float *DSO;

        # Parameters  of the Detector.
        int   nDetecU, nDetecV;
        float sDetecU, sDetecV;
        float dDetecU, dDetecV;
        float *offDetecU;
        float *offDetecV;
        float *DSD;
        float* dRoll;
        float* dPitch;
        float* dYaw;

        # The base unit we are working with in mm.
        float unitX;
        float unitY;
        float unitZ;

        #projection angle
        float alpha;
        float theta;
        float psi;
        # Centre of Rotation correction.
        float* COR;

        #Maximum length of cube
        float maxLength;

        #User option
        float accuracy;


    ctypedef struct Point3D:
        float x;
        float y;
        float z;

#TODO: Change from inline to stop duplication. Discussed on slack "python-questions" 28-Mar-2017
cdef inline void free_c_geometry(Geometry* c_geom):
    free(c_geom.offOrigX)
    free(c_geom.offOrigY)
    free(c_geom.offOrigZ)

    free(c_geom.offDetecU)
    free(c_geom.offDetecV)

    free(c_geom.dRoll)
    free(c_geom.dPitch)
    free(c_geom.dYaw)

    free(c_geom.COR)

    free(c_geom)

# Convert python-geometry to c-geometry.
# Interprets the size of np.array's in the C/cuda context.
# python-geometry's geometry.convert_contig_mode() is merged into this function. 
#TODO: Change from inline to stop duplication. Discussed on slack "python-questions" 28-Mar-2017
cdef inline Geometry* convert_to_c_geometry(p_geometry, int total_projections):
    cdef Geometry* c_geom =<Geometry *>malloc(sizeof(Geometry))

    ### Image ###
    c_geom.nVoxelX = p_geometry.nVoxel[2]
    c_geom.nVoxelY = p_geometry.nVoxel[1]
    c_geom.nVoxelZ = p_geometry.nVoxel[0]

    c_geom.sVoxelX = p_geometry.sVoxel[2]
    c_geom.sVoxelY = p_geometry.sVoxel[1]
    c_geom.sVoxelZ = p_geometry.sVoxel[0]
    
    c_geom.dVoxelX = p_geometry.dVoxel[2]
    c_geom.dVoxelY = p_geometry.dVoxel[1]
    c_geom.dVoxelZ = p_geometry.dVoxel[0]

    # TODO: array of constant for each alpha
    c_geom.offOrigX =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.offOrigX:
        raise MemoryError()
    c_geom.offOrigY =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.offOrigY:
        raise MemoryError()
    c_geom.offOrigZ =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.offOrigZ:
        raise MemoryError()
    c_geom.DSO =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.DSO:
        raise MemoryError()
    for i in range (total_projections):
        c_geom.offOrigX[i] = p_geometry.offOrigin[i][2]
        c_geom.offOrigY[i] = p_geometry.offOrigin[i][1]
        c_geom.offOrigZ[i] = p_geometry.offOrigin[i][0]
    for i in range(total_projections):
        c_geom.DSO[i] = p_geometry.DSO[i]

    ### Detector ###
    c_geom.nDetecU=p_geometry.nDetector[1]
    c_geom.nDetecV=p_geometry.nDetector[0]

    c_geom.sDetecU=p_geometry.sDetector[1]
    c_geom.sDetecV=p_geometry.sDetector[0]

    c_geom.dDetecU=p_geometry.dDetector[1]
    c_geom.dDetecV=p_geometry.dDetector[0]

    # TODO: array of constant for each alpha
    c_geom.offDetecU =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.offDetecU:
        raise MemoryError()
    c_geom.offDetecV =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.offDetecV:
        raise MemoryError()
    c_geom.DSD =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.DSD:
        raise MemoryError()
    for i in range (total_projections):
        c_geom.offDetecU[i] = p_geometry.offDetector[i][1]
        c_geom.offDetecV[i] = p_geometry.offDetector[i][0]
    for i in range(total_projections):
        c_geom.DSD[i] = p_geometry.DSD[i]

    # TODO: array of 0 for each alpha
    c_geom.dRoll =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.dRoll:
        raise MemoryError()
    c_geom.dPitch =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.dPitch:
        raise MemoryError()
    c_geom.dYaw =<float *>malloc(total_projections * sizeof(float))
    if not c_geom.dYaw:
        raise MemoryError()
    for i in range (total_projections):
        c_geom.dRoll[i] = p_geometry.rotDetector[i][2]
        c_geom.dPitch[i] = p_geometry.rotDetector[i][1]
        c_geom.dYaw[i] = p_geometry.rotDetector[i][0]

    # The base unit we are working with in mm.
    c_geom.unitX = 1
    c_geom.unitY = 1
    c_geom.unitZ = 1

    #projection angle
    # float alpha; #TODO: Check this is redundant

    # TODO: check this is correct line ~400 Ax.cpp. Seems like it can be omitted? Possibly redundant
    # Centre of Rotation correction.
    c_geom.COR =<float *>malloc(total_projections * sizeof(float))
    for i in range (total_projections):
        c_geom.COR[i] = p_geometry.COR[i]

    #Maximum length of cube
    # float maxLength; #TODO: Check this is redundant

    #User option
    c_geom.accuracy = p_geometry.accuracy

    return c_geom



from __future__ import division

import time
import copy
import numpy as np
import scipy as sp
import tigre
from tigre.algorithms.iterative_recon_alg import IterativeReconAlg
from tigre.algorithms.iterative_recon_alg import decorator
from tigre.utilities.Atb import Atb
from tigre.utilities.Ax import Ax
import tigre.algorithms as algs
import scipy.sparse.linalg


if hasattr(time, "perf_counter"):
    default_timer = time.perf_counter
else:
    default_timer = time.clock


class CGLS(IterativeReconAlg):  # noqa: D101
    __doc__ = (
        " CGLS solves the CBCT problem using the conjugate gradient least\n"
        " squares\n"
        " \n"
        "  CGLS(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    def initialize_algo(self):
        self.__r__ = self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        self.__p__ = Atb(self.__r__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)
        p_norm = np.linalg.norm(self.__p__.ravel(), 2)
        self.__gamma__ = p_norm * p_norm

    # Overide
    def run_main_iter(self):

        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        avgtime = []

        self.initialize_algo()
        for i in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(i)
            if self.Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)

            avgtic = default_timer()
            q = tigre.Ax(self.__p__, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
            q_norm = np.linalg.norm(q)
            alpha = self.__gamma__ / (q_norm * q_norm)
            self.res += alpha * self.__p__
            avgtoc = default_timer()
            avgtime.append(abs(avgtic - avgtoc))

            self.l2l[0, i] = np.linalg.norm(self.proj - tigre.Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids))
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                self.res -= alpha * self.__p__

                if self.verbose:
                    print("re-initilization of CGLS called at iteration:" + str(i))
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("CGLS exited due to divergence.")
                    return self.res
                self.re_init_at_iteration=i
                i=i-1
                self.initialize_algo()
                break

            self.__r__ -= alpha * q
            s = tigre.Atb(self.__r__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)
            s_norm = np.linalg.norm(s)

            gamma1 = s_norm * s_norm
            beta = gamma1 / self.__gamma__

            self.__gamma__ = gamma1
            self.__p__ = s + beta * self.__p__
            if self.Quameasopts is not None:
                self.error_measurement(res_prev, i)

        if self.verbose:
            print(
                "Average time taken for each iteration for CGLS:"
                + str(sum(avgtime) / len(avgtime))
                + "(s)"
            )

cgls = decorator(CGLS, name="cgls")

class LSQR(IterativeReconAlg): 
    __doc__ = (
        " LSQR solves the CBCT problem using the  least squares\n"
        "  LSQR(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    def initialize_algo(self):
        # Paige and Saunders //doi.org/10.1145/355984.355989

        # Enumeration as given in the paper for 'Algorithm LSQR'
        # (1) Initialize 
        self.__u__=self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        
        normr = np.linalg.norm(self.__u__.ravel(), 2)
        self.__u__ = self.__u__/normr

        self.__beta__ = normr
        self.__phibar__ = normr
        self.__v__ = Atb(self.__u__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)

        self.__alpha__ =np.linalg.norm(self.__v__.ravel(), 2)
        self.__v__ = self.__v__/self.__alpha__
        self.__rhobar__ = self.__alpha__
        self.__w__ = np.copy(self.__v__)

    def run_main_iter(self):
        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        avgtime = []
        self.initialize_algo()
        for i in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(i)
            if self.Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            avgtic = default_timer()    
            
            #% (3)(a)
            self.__u__ = tigre.Ax(self.__v__, self.geo, self.angles, "Siddon", gpuids=self.gpuids) - self.__alpha__*self.__u__
            self.__beta__ = np.linalg.norm(self.__u__.ravel(),2)
            self.__u__ = self.__u__ / self.__beta__
            
            #% (3)(b)
            self.__v__ = tigre.Atb(self.__u__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids) - self.__beta__*self.__v__
            self.__alpha__ = np.linalg.norm(self.__v__.ravel(),2)
            self.__v__ = self.__v__ / self.__alpha__    

            #% (4)(a-g)
            rho = np.sqrt(self.__rhobar__**2 + self.__beta__**2)
            c = self.__rhobar__ / rho
            s =  self.__beta__ / rho
            theta = s * self.__alpha__    
            self.__rhobar__ = - c * self.__alpha__    
            phi = c * self.__phibar__
            self.__phibar__ = s * self.__phibar__
            
            #% (5) Update x, w
            self.res = self.res + (phi / rho) * self.__w__
            self.__w__ = self.__v__ - (theta / rho) * self.__w__

            avgtoc = default_timer()
            avgtime.append(abs(avgtic - avgtoc))

            if self.Quameasopts is not None:
                self.error_measurement(res_prev, i)

            self.l2l[0, i] = np.linalg.norm(self.proj - tigre.Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids))
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                self.res -= (phi / rho) * (self.__v__-self.__w__)/((theta / rho))
                if self.verbose:
                    print("re-initilization of LSQR called at iteration:" + str(i))
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("LSQR exited due to divergence.")
                    return self.res
                self.re_init_at_iteration=i
                i=i-1
                self.initialize_algo()
                break

lsqr = decorator(LSQR, name="lsqr")

class hybrid_LSQR(IterativeReconAlg): 
    __doc__ = (
        " hybrid_LSQR solves the CBCT problem using the hybrid_LSQR\n"
        "  hybrid_LSQR(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
        self.__U__ = np.zeros((self.niter+1,np.prod(self.geo.nDetector)*len(self.angles)),dtype=np.float32)
        self.__V__ = np.zeros((self.niter,(np.prod(self.geo.nVoxel))),dtype=np.float32)
        self.__B__ = np.zeros((self.niter,self.niter+1),dtype=np.float32) #% Projected matrix
        self.__proj_rhs__ = np.zeros((self.niter+1,1),dtype=np.float32) #% Projected right hand side

    def initialize_algo(self):
        # Paige and Saunders //doi.org/10.1145/355984.355989
        # Enumeration as given in the paper for 'Algorithm LSQR'
        # % Initialise matrices

        # (1) Initialize 
        self.__u__=self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        
        normr = np.linalg.norm(self.__u__.ravel(), 2)
        self.__u__ = self.__u__/normr
        self.__U__[0]=self.__u__.ravel()

        self.__beta__ = normr
        self.__proj_rhs__[0]=normr


    def run_main_iter(self):
        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        avgtime = []
        self.initialize_algo()
        for i in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(i)
            
            avgtic = default_timer() 

            v = Atb(self.__u__,self.geo,self.angles,backprojection_type="matched",gpuids=self.gpuids)
            
            if i>0:
                v = np.reshape(v.ravel() - self.__beta__*self.__V__[i-1],v.shape)
    
            
            for j in range(i-1):
                v=np.reshape(v.ravel()-(self.__V__[j]*v.ravel())*self.__V__[j],v.shape)

   
            alpha = np.linalg.norm(v.ravel(), 2)
            v = v/alpha
            self.__V__[i] = v.ravel()

            #% Update U_{ii+1}
            self.__u__ = tigre.Ax(v, self.geo, self.angles, "Siddon", gpuids=self.gpuids) - alpha*self.__u__
            
            for j in range(i-1):
                self.__u__=np.reshape(self.__u__.ravel()-(self.__U__[j]*self.__u__.ravel())*self.__U__[j],self.__u__.shape)
                
            
            self.__beta__  = np.linalg.norm(self.__u__.ravel(), 2)
            self.__u__ = self.__u__ / self.__beta__ 
            self.__U__[i+1] = self.__u__.ravel()

            #% Update projected matrix
            self.__B__[i,i] = alpha
            self.__B__[i,i+1] = self.__beta__ 
            #% Malena. Proposed update: we should check algorithms breaks; 
            #% 'if abs(alpha) <= eps || abs(beta) <= eps' - end and save

            #% Solve the projected problem 
            #% (using the SVD of the small projected matrix)
            Bk = self.__B__[0:i+1,0:i+2]
            Uk, Sk, Vk = np.linalg.svd(np.transpose(Bk))
      
            if i==0:
                Sk = Sk[0]
            
            rhsk = self.__proj_rhs__[0:i+2]
            rhskhat = np.matmul(np.transpose(Uk),rhsk) #
            Dk = Sk**2 + self.lmbda**2

            rhskhat = Sk * rhskhat[0:i+1,0]
            yhat = rhskhat[0:i+1]/Dk
            y = np.matmul(np.transpose(Vk), yhat)

      


            self.l2l[0, i] = np.linalg.norm(self.proj - tigre.Ax(self.res + np.reshape(np.matmul(np.transpose(self.__V__[0:i+1]),y),self.res.shape), self.geo, self.angles, "Siddon", gpuids=self.gpuids))
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("hybrid LSQR exited due to divergence at iteration "+str(i))
                    return  self.res + np.reshape(np.matmul(np.transpose(self.__V__[0:i+1]),y),self.res.shape)
                
            #% Test for convergence. 
            #% msl: I still need to implement this. 
            #% msl: There are suggestions on the original paper. Let's talk about it!
        
        self.res = self.res + np.reshape(np.matmul(np.transpose(self.__V__),y),self.res.shape)
        return self.res
hybrid_lsqr = decorator(hybrid_LSQR, name="hybrid_lsqr")

class LSMR(IterativeReconAlg): 
    __doc__ = (
        " LSMR solves the CBCT problem using LSMR\n"
        "  LSMR(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    def initialize_algo(self):
        #% David Chin-Lung Fong and Michael Saunders //doi.org/10.1137/10079687X
        #% Enumeration as given in the paper for 'Algorithm LSMR'
        #% (1) Initialize 
        self.__u__=self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        normr = np.linalg.norm(self.__u__.ravel(), 2)
        self.__beta__ = normr
        self.__u__ = self.__u__/normr

        self.__v__ = Atb(self.__u__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)
        self.__alpha__ =np.linalg.norm(self.__v__.ravel(), 2)
        self.__v__ = self.__v__/self.__alpha__

        self.__alphabar__ = self.__alpha__
        self.__zetabar__ = self.__alpha__ * self.__beta__
        self.__rho__ = 1
        self.__rhobar__ = 1
        self.__cbar__ = 1
        self.__sbar__ = 0
        self.__h__ = self.__v__ 
        self.__hbar__ = 0

        #% Compute the residual norm ||r_k||
        self.__betadd__ = self.__beta__
        self.__betad__ = 0
        self.__rhod__ = 1
        self.__tautilda__ = 0
        self.__thetatilda__ = 0
        self.__zeta__ = 0
        self.__d__ = 0

    def run_main_iter(self):
        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        avgtime = []
        self.initialize_algo()
        for i in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(i)
            if self.Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            avgtic = default_timer()  
                    
            #% (3) Continue the bidiagonalization
            self.__u__ = tigre.Ax(self.__v__, self.geo, self.angles, "Siddon", gpuids=self.gpuids) - self.__alpha__*self.__u__
            self.__beta__ = np.linalg.norm(self.__u__.ravel(),2)
            self.__u__ = self.__u__ / self.__beta__
            
            self.__v__ = tigre.Atb(self.__u__, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids) - self.__beta__*self.__v__
            self.__alpha__ = np.linalg.norm(self.__v__.ravel(),2)
            self.__v__ = self.__v__ / self.__alpha__  

            #% (4) Construct and apply rotation \hat{P}_k
            alphahat = np.sqrt(self.__alphabar__**2 + self.lmbda**2)
            chat = self.__alphabar__/alphahat
            shat = self.lmbda/alphahat

            #% (5) Construct and apply rotation P_k
            rhopre = self.__rho__; 
            self.__rho__ = np.sqrt(alphahat**2 + self.__beta__**2)
            c = alphahat / self.__rho__
            s =  self.__beta__ / self.__rho__
            theta = s * self.__alpha__
            self.__alphabar__ = c * self.__alpha__

            #% (6) Construct and apply rotation \bar{P}_k
            thetabar = self.__sbar__  * self.__rho__
            rhobarpre = self.__rhobar__
            self.__rhobar__ = np.sqrt((self.__cbar__ *self.__rho__)**2 + theta**2)
            self.__cbar__ = self.__cbar__ * self.__rho__ / self.__rhobar__
            self.__sbar__ = theta / self.__rhobar__
            zetapre = self.__zeta__
            self.__zeta__ = self.__cbar__ * self.__zetabar__
            self.__zetabar__ = -self.__sbar__ * self.__zetabar__

            #% (7) Update \bar{h}, x, h
            self.__hbar__  = self.__h__ - (thetabar*self.__rho__/(rhopre*rhobarpre))*self.__hbar__ 
            self.res = self.res + (self.__zeta__ / (self.__rho__*self.__rhobar__)) * self.__hbar__ 
            self.__h__ = self.__v__ - (theta / self.__rho__) * self.__h__

            #% (8) Apply rotation \hat{P}_k, P_k
            betaacute = chat* self.__betadd__
            betacheck = - shat* self.__betadd__

            #% Computing ||r_k||

            betahat = c * betaacute
            betadd = -s * betaacute

            #% Update estimated quantities of interest.
            #%  (9) If k >= 2, construct and apply \tilde{P} to compute ||r_k||
            rhotilda = np.sqrt(self.__rhod__**2 + thetabar**2)
            ctilda = self.__rhod__ / rhotilda
            stilda = thetabar / rhotilda
            thetatildapre = self.__thetatilda__
            self.__thetatilda__ = stilda * self.__rhobar__
            self.__rhod__ = ctilda * self.__rhobar__
            #% betatilda = ctilda * betad + stilda * betahat; % msl: in the orinal paper, but not used
            self.__betad__ = -stilda * self.__betad__ + ctilda * betahat

            #% (10) Update \tilde{t}_k by forward substitution
            self.__tautilda__ = (zetapre - thetatildapre* self.__tautilda__) / rhotilda
            taud = (self.__zeta__ - self.__thetatilda__*self.__tautilda__) / self.__rhod__
            
            #% (11) Compute ||r_k||
            self.__d__ = self.__d__ + betacheck**2
            gamma_var = self.__d__ + (self.__betad__ - taud)**2 + betadd**2

            avgtoc = default_timer()
            avgtime.append(abs(avgtic - avgtoc))

            if self.Quameasopts is not None:
                self.error_measurement(res_prev, i)

            self.l2l[0, i] = np.linalg.norm(self.proj - tigre.Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids))
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                self.res -= (self.__zeta__ / (self.__rho__*self.__rhobar__)) * self.__hbar__ 
                if self.verbose:
                    print("re-initilization of LSMR called at iteration:" + str(i))
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("LSMR exited due to divergence.")
                    return self.res
                self.re_init_at_iteration=iter
                iter=iter-1
                self.initialize_algo()
                break
lsmr = decorator(LSMR, name="lsmr")

class IRN_TV_CGLS(IterativeReconAlg):
    __doc__ = (
        " IRN_TV_CGLS solves the CBCT problem using CGLS with TV constraints\n"
        "  IRN_TV_CGLS(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs):
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    
    def __build_weights__(self):
        Dxx=np.copy(self.res)
        Dyx=np.copy(self.res)
        Dzx=np.copy(self.res)

        Dxx[0:-2,:,:]=self.res[0:-2,:,:]-self.res[1:-1,:,:]
        Dyx[:,0:-2,:]=self.res[:,0:-2,:]-self.res[:,1:-1,:]
        Dzx[:,:,0:-2]=self.res[:,:,0:-2]-self.res[:,:,1:-1]
 
        return (Dxx**2+Dyx**2+Dzx**2+1e-6)**(-1/4)

    def Lx(self,W,img):
        Dxx=np.copy(img)
        Dyx=np.copy(img)
        Dzx=np.copy(img)

        Dxx[0:-2,:,:]=img[0:-2,:,:]-img[1:-1,:,:]
        Dyx[:,0:-2,:]=img[:,0:-2,:]-img[:,1:-1,:]
        Dzx[:,:,0:-2]=img[:,:,0:-2]-img[:,:,1:-1]

        return np.stack((W*Dxx,W*Dyx,W*Dzx),axis=0)
        
    def Ltx(self,W,img3):
        Wx_1 = W * img3[0,:,:,:]
        Wx_2 = W * img3[1,:,:,:]
        Wx_3 = W * img3[2,:,:,:]

        DxtWx_1=Wx_1
        DytWx_2=Wx_2
        DztWx_3=Wx_3
        
        DxtWx_1[1:-2,:,:]=Wx_1[1:-2,:,:]-Wx_1[0:-3,:,:]
        DxtWx_1[-1,:,:]=-Wx_1[-2,:,:]
        
        DytWx_2[:,1:-2,:]=Wx_2[:,1:-2,:]-Wx_2[:,0:-3,:]
        DytWx_2[:,-1,:]=-Wx_2[:,-2,:]
        
        DztWx_3[:,:,1:-2]=Wx_3[:,:,1:-2]-Wx_3[:,:,0:-3]
        DztWx_3[:,:,-1]=-Wx_3[:,:,-2]

        return DxtWx_1 + DytWx_2 + DztWx_3

    def run_main_iter(self):
        self.l2l = np.zeros((1, self.niter*self.niter_outer), dtype=np.float32)
        avgtime = []

        res0=self.res
   
        for outer in range(self.niter_outer):
            if self.verbose:
                niter=self.niter
                self.niter=self.niter_outer
                self._estimate_time_until_completion(outer)
                self.niter=niter
            if self.Quameasopts is not None:
                res_prev = copy.deepcopy(self.res)
            avgtic = default_timer()    


            W=self.__build_weights__()
            self.res=res0
    
            prox_aux_1 =Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
            prox_aux_2 = self.Lx(W,self.res)*np.sqrt(self.lmbda)
    
            r_aux_1 = self.proj - prox_aux_1
            r_aux_2 = -prox_aux_2
            #% Malena: changed the format, r_aux_2 is 3
            #% r = cat(3,r_aux_1, r_aux_2); % Malena: size guide, erase later, N x N x (100 + N-1)
            p_aux_1 = tigre.Atb(r_aux_1, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)
            p_aux_2 = np.sqrt(self.lmbda)*self.Ltx(W, r_aux_2)
            p = p_aux_1 + p_aux_2

            gamma=np.linalg.norm(p.ravel(),2)**2

            for i in range(self.niter):
                res0=self.res

                q_aux_1 = tigre.Ax(p, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
                q_aux_2 = self.Lx(W,p)*np.sqrt(self.lmbda)

                #% q = cat(3, q_aux_1, q_aux_2{1},q_aux_2{2},q_aux_2{3}); % Probably never need to actually do this
                #% alpha=gamma/norm(q(:),2)^2;
                alpha=gamma/(np.linalg.norm(q_aux_1.ravel(),2)**2 + np.linalg.norm(q_aux_2[0].ravel(),2)**2 + np.linalg.norm(q_aux_2[1].ravel(),2)**2+np.linalg.norm(q_aux_2[2].ravel(),2)**2)
                self.res=self.res+alpha*p
                aux=self.proj-tigre.Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
                #% residual norm or the original least squares (not Tikhonov).
                #% Think if that is what we want of the NE residual
                self.l2l[0, outer*self.niter+i] = np.linalg.norm(aux.ravel(),2)

              
                #% If step is adecuate, then continue withg CGLS
                r_aux_1 = r_aux_1-alpha*q_aux_1
                r_aux_2=r_aux_2-alpha*q_aux_2

                s_aux_1 = tigre.Atb(r_aux_1, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)
                s_aux_2 =  np.sqrt(self.lmbda) * self.Ltx(W, r_aux_2)
                s = s_aux_1 + s_aux_2

                gamma1=np.linalg.norm(s.ravel(),2)**2
                beta=gamma1/gamma
                gamma=gamma1
                p=s+beta*p

            avgtoc = default_timer()
            avgtime.append(abs(avgtic - avgtoc))

            if self.Quameasopts is not None:
                self.error_measurement(res_prev, outer)

irn_tv_cgls = decorator(IRN_TV_CGLS, name="irn_tv_cgls")


class AB_GMRES(IterativeReconAlg): 
    __doc__ = (
        " AB_GMRES solves the CBCT problem using preconditioned GMRES\n"
        "  AB_GMRES(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
            # Don't precompute V and W.
            kwargs.update(dict(W=None, V=None))
            kwargs.update(dict(blocksize=angles.shape[0]))
            self.re_init_at_iteration = 0
            if "backprojector" in kwargs:
                backproject=kwargs.pop("backprojector")
            else:
                backproject="matched"
            if backproject == "matched":
                self.backproject=Atb
            elif backproject == "FDK":
                self.backproject=algs.fdk
            IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
      

    def __compute_res__(self,x,w,y):
        y=y.astype(np.float32)
        for i in range(w.shape[0]):
            x=x+self.backproject(np.reshape(w[i],self.proj.shape),self.geo,self.angles,gpuids=self.gpuids)*y[i]
        return x

    def run_main_iter(self):

        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        w = np.zeros((self.niter+1,np.prod(self.geo.nDetector)*len(self.angles)),dtype=np.float32)
        r=self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        w[0] = r.ravel()/np.linalg.norm(r.ravel(), 2)
        h=np.zeros((self.niter,self.niter+1),dtype=np.float32)
        for k in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(k)
                
            qk=Ax(self.backproject(np.reshape(w[k],self.proj.shape),self.geo,self.angles,gpuids=self.gpuids),self.geo, self.angles, "Siddon", gpuids=self.gpuids)
            e1=np.zeros(k+2)
            e1[0]=1
            for i in range(k+1):
                h[k,i]=sum(qk.ravel()*w[i])
                qk=qk.ravel()-h[k,i]*w[i]
            
            h[k,k+1]=np.linalg.norm(qk.ravel(),2)
            w[k+1]=qk.ravel()/h[k,k+1]
            y=np.linalg.lstsq(np.transpose(h[0:k+1,0:k+2]),e1*np.linalg.norm(r.ravel(),2),rcond=None)
            y=y[0]
            self.l2l[0, i] = np.linalg.norm((self.proj - tigre.Ax(self.__compute_res__(self.res,w[0:k+1],y),self.geo,self.angles, "Siddon",gpuids=self.gpuids)).ravel(),2)
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("AB-GMRES exited due to divergence at iteration "+str(i))
                    return  self.__compute_res__(self.res,w[0:k+1],y)
             
        self.res=self.__compute_res__(self.res,w[0:-1],y)
        return self.res
ab_gmres = decorator(AB_GMRES, name="ab_gmres")



class BA_GMRES(IterativeReconAlg): 
    __doc__ = (
        " BA_GMRES solves the CBCT problem using preconditioned GMRES\n"
        "  AB_GMRES(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
            # Don't precompute V and W.
            kwargs.update(dict(W=None, V=None))
            kwargs.update(dict(blocksize=angles.shape[0]))
            self.re_init_at_iteration = 0
            if "backprojector" in kwargs:
                backproject=kwargs.pop("backprojector")
            else:
                backproject="matched"
            if backproject == "matched":
                self.backproject=Atb
            elif backproject == "FDK":
                self.backproject=algs.fdk
            IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)

    def __compute_res__(self,x,w,y):
        y=y.astype(np.float32)
        for i in range(w.shape[0]):
            x=x+np.reshape(w[i],self.res.shape)*y[i]
        return x

    def run_main_iter(self):

        self.l2l = np.zeros((1, self.niter), dtype=np.float32)
        w = np.zeros((self.niter+1,(np.prod(self.geo.nVoxel))),dtype=np.float32)
        r=self.backproject(self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids), self.geo, self.angles, gpuids=self.gpuids)
        w[0] = r.ravel()/np.linalg.norm(r.ravel(), 2)
        h=np.zeros((self.niter,self.niter+1),dtype=np.float32)

        for k in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(k)
                
            qk=self.backproject(Ax(np.reshape(w[k],self.res.shape),self.geo,self.angles, "Siddon",gpuids=self.gpuids),self.geo, self.angles, gpuids=self.gpuids)
            e1=np.zeros(k+2)
            e1[0]=1
            for i in range(k+1):
                h[k,i]=sum(qk.ravel()*w[i])
                qk=qk.ravel()-h[k,i]*w[i]
            
            h[k,k+1]=np.linalg.norm(qk.ravel(),2)
            w[k+1]=qk.ravel()/h[k,k+1]
            y=np.linalg.lstsq(np.transpose(h[0:k+1,0:k+2]),e1*np.linalg.norm(r.ravel(),2),rcond=None)
            y=y[0]

            self.l2l[0, i] = np.linalg.norm((self.proj - tigre.Ax(self.__compute_res__(self.res,w[0:k+1],y), self.geo, self.angles, "Siddon", gpuids=self.gpuids)).ravel(),2)
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("BA-GMRES exited due to divergence at iteration "+str(i))
                    return  self.__compute_res__(self.res,w[0:k+1],y)
             
        self.res=self.__compute_res__(self.res,w[0:-1],y)
        return self.res
ba_gmres = decorator(BA_GMRES, name="ba_gmres")


class hybrid_fLSQR_TV(IterativeReconAlg): 
    __doc__ = (
        " hybrid_fLSQR_TV solves the CBCT problem using preconditioned hybrid flexuble LSQR with TV regularization\n"
        "  AB_GMRES(PROJ,GEO,ANGLES,NITER) solves the reconstruction problem\n"
        "  using the projection data PROJ taken over ALPHA angles, corresponding\n"
        "  to the geometry descrived in GEO, using NITER iterations."
    ) + IterativeReconAlg.__doc__

    def __init__(self, proj, geo, angles, niter, **kwargs): 
        # Don't precompute V and W.
        kwargs.update(dict(W=None, V=None))
        kwargs.update(dict(blocksize=angles.shape[0]))
        self.re_init_at_iteration = 0
        IterativeReconAlg.__init__(self, proj, geo, angles, niter, **kwargs)
        self.__U__ = np.zeros((self.niter+1,np.prod(self.geo.nDetector)*len(self.angles)),dtype=np.float32)
        self.__V__ = np.zeros((self.niter,(np.prod(self.geo.nVoxel))),dtype=np.float32)
        self.__Z__ = np.zeros((self.niter,(np.prod(self.geo.nVoxel))),dtype=np.float32)

        self.__M__ = np.zeros((self.niter,self.niter+1),dtype=np.float32) #% Projected matrix
        self.__T__ = np.zeros((self.niter,self.niter),dtype=np.float32) #% Projected matrix
        self.__proj_rhs__ = np.zeros((self.niter+1,1),dtype=np.float32) #% Projected right hand side
    

      
    def __build_weights__(self):
        Dxx=np.copy(self.res)
        Dyx=np.copy(self.res)
        Dzx=np.copy(self.res)

        Dxx[0:-2,:,:]=self.res[0:-2,:,:]-self.res[1:-1,:,:]
        Dyx[:,0:-2,:]=self.res[:,0:-2,:]-self.res[:,1:-1,:]
        Dzx[:,:,0:-2]=self.res[:,:,0:-2]-self.res[:,:,1:-1]
 
        return (Dxx**2+Dyx**2+Dzx**2+1e-6)**(-1/4)

    def Lx(self,W,img):
        img=np.reshape(img,self.res.shape)
        Dxx=np.copy(img)
        Dyx=np.copy(img)
        Dzx=np.copy(img)

        Dxx[0:-2,:,:]=img[0:-2,:,:]-img[1:-1,:,:]
        Dyx[:,0:-2,:]=img[:,0:-2,:]-img[:,1:-1,:]
        Dzx[:,:,0:-2]=img[:,:,0:-2]-img[:,:,1:-1]

        return np.stack((W*Dxx,W*Dyx,W*Dzx),axis=0)

    def Ltx(self,W,img3):
        img3 =np.reshape(img3,(3,*self.res.shape))
        Wx_1 = W * img3[0,:,:,:]
        Wx_2 = W * img3[1,:,:,:]
        Wx_3 = W * img3[2,:,:,:]

        DxtWx_1=Wx_1
        DytWx_2=Wx_2
        DztWx_3=Wx_3
        
        DxtWx_1[1:-2,:,:]=Wx_1[1:-2,:,:]-Wx_1[0:-3,:,:]
        DxtWx_1[-1,:,:]=-Wx_1[-2,:,:]
        
        DytWx_2[:,1:-2,:]=Wx_2[:,1:-2,:]-Wx_2[:,0:-3,:]
        DytWx_2[:,-1,:]=-Wx_2[:,-2,:]
        
        DztWx_3[:,:,1:-2]=Wx_3[:,:,1:-2]-Wx_3[:,:,0:-3]
        DztWx_3[:,:,-1]=-Wx_3[:,:,-2]

        return DxtWx_1 + DytWx_2 + DztWx_3

    def mvT(self,k_aux,x ):
            return x.ravel() - k_aux.ravel()*np.sum(x.ravel())
    def mv(self,k_aux,x ):
            return x.ravel() - np.dot(k_aux.ravel(),x.ravel())
        
    def run_main_iter(self):
        # % Initialise matrices

        # (1) Initialize 
        u=self.proj - Ax(self.res, self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        
        k_aux = Ax(np.ones(self.res.shape,dtype=np.float32)/np.sqrt(np.prod(self.geo.nVoxel)), self.geo, self.angles, "Siddon", gpuids=self.gpuids)
        xA0 = 1/(np.sqrt(np.prod(self.geo.nVoxel))*np.linalg.norm(k_aux.ravel(),2)**2)*np.dot(k_aux.ravel(),u.ravel())
        xA0 =np.ones(self.res.shape,dtype=np.float32)*xA0
        
        k_aux = 1/(np.sqrt(np.prod(self.geo.nVoxel))*np.linalg.norm(k_aux.ravel(),2)**2)*Atb(k_aux, self.geo, self.angles, backprojection_type="matched", gpuids=self.gpuids)

        u = u - Ax(xA0, self.geo, self.angles, "Siddon", gpuids=self.gpuids)

        normr=np.linalg.norm(u.ravel(),2)
        u=u/normr
        self.__U__[0]=u.ravel()

        self.__proj_rhs__[0]=normr

        if np.max(self.res)==0:
            W=np.ones(self.res.shape,dtype=np.float32)
        else:
            W=self.__build_weights__()

        self.l2l = np.zeros((1, self.niter), dtype=np.float32)

        for i in range(self.niter):
            if self.verbose:
                self._estimate_time_until_completion(i)
            v=Atb(u,self.geo,self.angles,backprojection_type="matched", gpuids=self.gpuids)
            for j in range(i+1):
                self.__T__[i,j] = np.dot(self.__V__[j],v.ravel())
                v = v.ravel() - self.__T__[i,j]*self.__V__[j]
            v=np.reshape(v,self.res.shape)
            self.__T__[i,i] = np.linalg.norm(v.ravel(),2)
            v=v/self.__T__[i,i]

            z=self.mvT(k_aux,v)

            L = sp.sparse.linalg.LinearOperator((np.prod(self.res.shape),np.prod(self.res.shape)*3), matvec=lambda x: self.Ltx(W,x).ravel(),rmatvec=lambda x: self.Lx(W,x).ravel())
            aux_z=sp.sparse.linalg.lsqr(L,z.ravel(),iter_lim=50)
            L = sp.sparse.linalg.LinearOperator((np.prod(self.res.shape)*3,np.prod(self.res.shape)), matvec=lambda x: self.Lx(W,x).ravel(),rmatvec=lambda x: self.Ltx(W,x).ravel())
            z=sp.sparse.linalg.lsqr(L,aux_z[0].ravel(),iter_lim=50)
            z=z[0].astype(np.float32)

            z=self.mv(k_aux,z)

            self.__V__[i]=v.ravel()
            self.__Z__[i]=z.ravel()

            # Update U and projected matrix M
            u = Ax(np.reshape(z,self.res.shape), self.geo, self.angles, "Siddon", gpuids=self.gpuids)
            for j in range(i+1):
                self.__M__[i,j] = np.dot(self.__U__[j],u.ravel())
                u = u.ravel() - np.dot(self.__M__[i,j],self.__U__[j])
            self.__M__[i,i+1]=np.linalg.norm(u.ravel(),2)
            u=u/self.__M__[i,i+1]
            u=np.reshape(u,self.proj.shape)
            self.__U__[i+1]=u.ravel()

            ### Solve the regularized projected problem
            # (using the DVF of the small projected matrix)

            Mk=self.__M__[0:i+1,0:i+2]
            # Prepare the projected regularization term
            WZ=np.zeros((i+1,(3*np.prod(self.geo.nVoxel))),dtype=np.float32)
            for j in range(i+1):
                # This can be more efficiently...
                # DZ can be saved and updated at each iteration
                WZ[j]=self.Lx(W,self.__Z__[j]).ravel()
            __,ZRk =sp.linalg.qr(np.transpose(WZ),mode='economic')
            ZRksq=ZRk[0:i+1,0:i+1]
            rhsk=self.__proj_rhs__[0:i+2]
            MZk=np.concatenate((np.transpose(Mk),self.lmbda*ZRksq))
            rhsZk=np.concatenate((rhsk,np.zeros((i+1,1),dtype=np.float32)))
            y = np.linalg.lstsq(MZk, rhsZk,rcond=None)
            print(y[0])
            d=np.matmul(np.transpose(self.__Z__[0:i+1]),y[0])
            
            x = self.res + np.reshape(d,self.res.shape) + xA0
            self.l2l[0, i] = np.linalg.norm((self.proj - tigre.Ax(x, self.geo, self.angles, "Siddon", gpuids=self.gpuids)).ravel(),2)
            if i > 0 and self.l2l[0, i] > self.l2l[0, i - 1]:
                if self.re_init_at_iteration + 1 == i or not self.restart:
                    print("BA-GMRES exited due to divergence at iteration "+str(i))
                    return  x
             
        return x

hybrid_flsqr_tv = decorator(hybrid_fLSQR_TV, name="hybrid_flsqr_tv")



class Error(Exception):
    """Base class for exceptions in this module"""

    def __init__(self):
        Exception.__init__(self)


class TigreCudaCallError(Error):
    """Error list indexed against values in errors.hpp."""

    def __init__(self, txt, error_index):
        Error.__init__(self)
        self.error_list = [
            "GPU: Cuda error with exit code 1.",
            "There are no available device(s) that support CUDA",
            "GPU: One (or more) of your GPUs is being heavily used by another program (possibly graphics-based).\n"  # noqa: E501
            "Free the GPU to run TIGRE",
            "GPU: Assertion Failed. Logic behind spliting flawed! Please tell: ander.biguri@gmail.com",  # noqa: E501
            "GPU: Assertion Failed. Memory needed calculation broken! Please tell: ander.biguri@gmail.com",  # noqa: E501
        ]
        self.error_index = error_index
        self.txt = txt

    def __str__(self):
        return "\n".join([self.txt, self.error_list[self.error_index - 1]])



#   This file is part of the TIGRE Toolbox

#   Copyright (c) 2015, University of Bath and
#                       CERN-European Organization for Nuclear Research
#                       All rights reserved.

#   License:            Open Source under BSD.
#                       See the full license at
#                       https://github.com/CERN/TIGRE/license.txt

#   Contact:            tigre.toolbox@gmail.com
#   Codes:              https://github.com/CERN/TIGRE/
# --------------------------------------------------------------------------
#   Coded by:          Tomoyuki SADAKANE

cimport numpy as np 
import numpy as np
from tigre.utilities.errors import TigreCudaCallError
from tigre.utilities.cuda_interface._gpuUtils cimport GpuIds as c_GpuIds, convert_to_c_gpuids, free_c_gpuids

np.import_array()

from libc.stdlib cimport malloc, free 

cdef extern from "numpy/arrayobject.h":
    void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)
    void PyArray_CLEARFLAGS(np.ndarray arr, int flags)

cdef extern from "RandomNumberGenerator.hpp":
    cdef void poisson_1d(float* img, size_t uiLen, float* dst, c_GpuIds gpuids)
    cdef void poisson_gaussian_1d(float* img, size_t uiLen, float fGaussMu, float fGaussSigma, float* dst, c_GpuIds gpuids)

def cuda_raise_errors(error_code):
    if error_code:
        raise TigreCudaCallError('RandomNumberGenerator::', error_code)

def add_poisson(np.ndarray[np.float32_t, ndim=3] src, gpuids=None):
    # print("add_poisson()")
    cdef c_GpuIds* c_gpuids = convert_to_c_gpuids(gpuids)
    if not c_gpuids:
        raise MemoryError()

    cdef np.npy_intp size_img[3]
    size_img[0]= <np.npy_intp> src.shape[0]
    size_img[1]= <np.npy_intp> src.shape[1]
    size_img[2]= <np.npy_intp> src.shape[2]

    cdef float* c_imgout = <float*> malloc(size_img[0] *size_img[1] *size_img[2]* sizeof(float))

    cdef long imgsize[3]
    imgsize[0] = <long> size_img[2]
    imgsize[1] = <long> size_img[1]
    imgsize[2] = <long> size_img[0]

    src = np.ascontiguousarray(src)

    cdef float* c_src = <float*> src.data
    cdef np.npy_intp c_uiSigLen = <np.npy_intp> (size_img[0] *size_img[1] *size_img[2])
    cuda_raise_errors(poisson_1d(c_src, c_uiSigLen, c_imgout, c_gpuids[0]))
    imgout = np.PyArray_SimpleNewFromData(3, size_img, np.NPY_FLOAT32, c_imgout)
    PyArray_ENABLEFLAGS(imgout, np.NPY_OWNDATA)
    
    return imgout

def add_noise(np.ndarray[np.float32_t, ndim=3] poisson_lambda,
              np.float32_t gaussian_mu,
              np.float32_t gaussian_sigma,
              gpuids=None):
    # print("add_noise()")
    cdef c_GpuIds* c_gpuids = convert_to_c_gpuids(gpuids)
    if not c_gpuids:
        raise MemoryError()

    cdef np.npy_intp size_img[3]
    size_img[0]= <np.npy_intp> poisson_lambda.shape[0]
    size_img[1]= <np.npy_intp> poisson_lambda.shape[1]
    size_img[2]= <np.npy_intp> poisson_lambda.shape[2]

    cdef float* c_imgout = <float*> malloc(size_img[0] *size_img[1] *size_img[2]* sizeof(float))

    cdef long imgsize[3]
    imgsize[0] = <long> size_img[2]
    imgsize[1] = <long> size_img[1]
    imgsize[2] = <long> size_img[0]

    poisson_lambda = np.ascontiguousarray(poisson_lambda)

    cdef float* c_src = <float*> poisson_lambda.data
    cdef np.npy_intp c_uiSigLen = <np.npy_intp> (size_img[0] *size_img[1] *size_img[2])
    cdef np.float32_t c_fGaussMu = gaussian_mu
    cdef np.float32_t c_fGaussSigma = gaussian_sigma
    cuda_raise_errors(poisson_gaussian_1d(c_src, c_uiSigLen, c_fGaussMu, c_fGaussSigma, c_imgout, c_gpuids[0]))
    imgout = np.PyArray_SimpleNewFromData(3, size_img, np.NPY_FLOAT32, c_imgout)
    PyArray_ENABLEFLAGS(imgout, np.NPY_OWNDATA)
    
    return imgout



""" Visualization of ordered subset of angles
#   This file is part of the TIGRE Toolbox

#   Copyright (c) 2015, University of Bath and
#                       CERN-European Organization for Nuclear Research
#                       All rights reserved.

#   License:            Open Source under BSD.
#                       See the full license at
#                       https://github.com/CERN/TIGRE/license.txt

#   Contact:            tigre.toolbox@gmail.com
#   Codes:              https://github.com/CERN/TIGRE/
# --------------------------------------------------------------------------
#   Coded by:          Tomoyuki SADAKANE
"""

import numpy as np
from matplotlib import pyplot as plt
import matplotlib.animation as animation

class PlotAngles:
    """Generate ordered subset angles animation
    Parameters
    ------------
        angle_blocks: List of subsets of extended (alpha, theta, psi) angles.
        angle_index: List of subsets of indeces of angles.
        savegif: Specify file name to create a gif file.
        show_plot: Set False not to display the plot.
        figsize: figsize to be passed to plt.figure()
        interval: Animation interval of the frames in ms.
        draw_number: Draw number in the plot. Default is True.
        size_marker: Size of the angle marker. Default is None to use the default of plt.scatter().
        angles_extended: If angles are not extended, set False. Default is True.
    """
    def __init__(self,
                angle_blocks,
                angle_index,
                savegif=None,
                show_plot=True,
                figsize=None,
                interval=500,
                draw_number=True,
                size_marker=None,
                angles_extended=True):
        if not angles_extended:
            if type(angle_blocks) is list:
                for idx, block in enumerate(angle_blocks):
                    a1 = block
                    a2 = np.zeros(a1.shape[0], dtype=np.float32)
                    angle_blocks[idx] = np.vstack((a1, a2, a2)).T
            elif type(angle_blocks) is np.ndarray:
                if angle_blocks.ndim == 1:
                    # treat as blocksize = 1
                    a1 = angle_blocks
                    a2 = np.zeros(a1.shape[0], dtype=np.float32)
                    angle_blocks = np.vstack((a1, a2, a2)).T
                    angle_blocks = list(angle_blocks)

        if type(angle_blocks) is np.ndarray:
            angle_blocks = [angle_blocks]
        if type(angle_index) is np.ndarray:
            angle_index = [angle_index]
        if len(angle_blocks) == 0 or len(angle_index) == 0:
            raise TypeError("angle_blocks and angle_index must be a lists of sizes greater than 0.")

        if angle_blocks[0].ndim != angle_index[0].ndim:
            # angles is extended
            if angle_blocks[0].ndim == 1:
                self.angle_blocks = [np.array([angle_blocks[i][0]]) for i in range(len(angle_blocks))]
            else:
                self.angle_blocks = [angle_blocks[i][:, 0] for i in range(len(angle_blocks))]
        else:
            self.angle_blocks = angle_blocks

        self.angle_index = angle_index

        self.show_plot  = show_plot
        self.figsize = figsize
        self.savegif = savegif

        self.interval = interval
        self.draw_number = draw_number
        self.size_marker = size_marker

        self._run()

    def _update_frame(self, it, fig, radius):
        fig.clf()
        axis = fig.add_subplot(1, 1, 1)
        rangewidth = radius*1.1
        axis.set_aspect("equal")
        axis.set_xlim([-rangewidth, +rangewidth]*radius)
        axis.set_ylim([-rangewidth, +rangewidth]*radius)
        axis.axhline(0, ls="solid", color="gray", linewidth = 0.5)
        axis.axvline(0, ls="solid", color="gray", linewidth = 0.5)
        axis.set_xlabel("x")
        axis.set_ylabel("y")
        radius_text = radius *0.9
        block_count = len(self.angle_blocks)
        for idx in range(it):
            if idx < block_count:
                for idx_block in range(self.angle_blocks[idx].shape[0]):
                    x = radius*np.cos(self.angle_blocks[idx][idx_block])
                    y = radius*np.sin(self.angle_blocks[idx][idx_block])
                    axis.scatter(x, y, color="red", s=self.size_marker)
                    if self.draw_number:
                        x = radius_text*np.cos(self.angle_blocks[idx][idx_block])
                        y = radius_text*np.sin(self.angle_blocks[idx][idx_block])
                        axis.text(x, y, str(idx), verticalalignment = "center", horizontalalignment = "center")

        if it < block_count:
            for idx_block in range(self.angle_blocks[it].shape[0]):
                self.x_old = radius*np.cos(self.angle_blocks[it][idx_block])
                self.y_old = radius*np.sin(self.angle_blocks[it][idx_block])
                axis.scatter(self.x_old, self.y_old, color="green", s=self.size_marker)
                if self.draw_number:
                    x = radius_text*np.cos(self.angle_blocks[it][idx_block])
                    y = radius_text*np.sin(self.angle_blocks[it][idx_block])
                    axis.text(x, y, str(it), verticalalignment = "center", horizontalalignment = "center")


    def _run(self):

        radius = 1
        self.x_old = radius*np.cos(self.angle_blocks[0][0])
        self.y_old = radius*np.sin(self.angle_blocks[0][0])

        fig = plt.figure(figsize=self.figsize)
        ani = animation.FuncAnimation(
                            fig,
                            self._update_frame,
                            fargs=(fig, radius),
                            interval=self.interval,
                            repeat_delay=1000,
                            frames = len(self.angle_blocks)+1)
        if self.savegif is not None:
            ani.save(self.savegif, writer="pillow")
            self._show()
        else:
            self._show()

    def _show(self):
        if self.show_plot:
            plt.show()

plot_angles = PlotAngles



from __future__ import division

import numpy as np
from tigre.utilities.geometry import Geometry


class ConeGeometryDefault(Geometry):
    def __init__(self, high_resolution=True, nVoxel=None):

        Geometry.__init__(self)
        if high_resolution:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536.0  # Distance Source Detector      (mm)
            self.DSO = 1000.0  # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((512, 512))  # number of pixels              (px)
            self.dDetector = np.array((0.8, 0.8))  # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector  # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((512, 512, 512))  # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))  # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel  # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))  # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))  # Offset of Detector            (mm)

            # Auxiliary
            self.accuracy = 0.5  # Accuracy of FWD proj          (vx/sample)  # noqa: E501
            # Mode
            self.mode = "cone"  # parallel, cone                ...
            self.filter = None
        else:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536.0  # Distance Source Detector      (mm)
            self.DSO = 1000.0  # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((128, 128))  # (V,U) number of pixels        (px)
            self.dDetector = np.array((0.8, 0.8)) * 4  # size of each pixel            (mm)
            self.sDetector = self.nDetector * self.dDetector  # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((64, 64, 64))  # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))  # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel  # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))  # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))  # Offset of Detector            (mm)
            self.rotDetector = np.array((0, 0, 0))
            # Auxiliary
            self.accuracy = 0.5  # Accuracy of FWD proj          (vx/sample)  # noqa: E501
            # Mode
            self.mode = "cone"  # parallel, cone                ...
            self.filter = None
        if nVoxel is not None:
            # VARIABLE                                          DESCRIPTION                    UNITS
            # -------------------------------------------------------------------------------------
            self.DSD = 1536.0  # Distance Source Detector      (mm)
            self.DSO = 1000.0  # Distance Source Origin        (mm)
            # Detector parameters
            self.nDetector = np.array((nVoxel[1], nVoxel[2]))  # (V,U) number of pixels        (px)
            self.dDetector = np.array([0.8, 0.8])  # size of each pixel            (mm)
            self.sDetector = self.dDetector * self.nDetector  # total size of the detector    (mm)
            # Image parameters
            self.nVoxel = np.array((nVoxel))  # number of voxels              (vx)
            self.sVoxel = np.array((256, 256, 256))  # total size of the image       (mm)
            self.dVoxel = self.sVoxel / self.nVoxel  # size of each voxel            (mm)
            # Offsets
            self.offOrigin = np.array((0, 0, 0))  # Offset of image from origin   (mm)
            self.offDetector = np.array((0, 0))  # Offset of Detector            (mm)
            self.rotDetector = np.array((0, 0, 0))
            # Auxiliary
            self.accuracy = 0.5  # Accuracy of FWD proj          (vx/sample)  # noqa: E501
            # Mode
            self.mode = "cone"  # parallel, cone



from __future__ import division
import os
import numpy as np
import scipy.io
import scipy.ndimage.interpolation


def load_head_phantom(number_of_voxels=None):
    if number_of_voxels is None:
        number_of_voxels = np.array((128, 128, 128))
    dirname = os.path.dirname(__file__)
    dirname = os.path.join(dirname, "../../../Common/data/head.mat")
    if not os.path.isfile(dirname):
        dirname = os.path.dirname(__file__)
        dirname = os.path.join(dirname, "./../../data/head.mat")
    test_data = scipy.io.loadmat(dirname)

    # Loads data in F_CONTIGUOUS MODE (column major), convert to Row major
    image = test_data["img"].transpose(2, 1, 0).copy()
    image_dimensions = image.shape

    zoom_x = number_of_voxels[0] / image_dimensions[0]
    zoom_y = number_of_voxels[1] / image_dimensions[1]
    zoom_z = number_of_voxels[2] / image_dimensions[2]

    # TODO: add test for this is resizing and not simply zooming
    resized_image = scipy.ndimage.interpolation.zoom(
        image, (zoom_x, zoom_y, zoom_z), order=3, prefilter=False
    )

    return resized_image



import numpy as np
from _tvdenoising import tvdenoise

from .gpu import GpuIds


def im3ddenoise(img, iter=50, lmbda=15.0, gpuids=None):
    imgmin = np.amin(img.ravel())
    img = img - imgmin
    imgmax = np.amax(img.ravel())
    img = img / imgmax

    if gpuids is None:
        gpuids = GpuIds()

    img = tvdenoise(img, iter, lmbda, gpuids)

    img *= imgmax
    img += imgmin

    return img



from libc.stdlib cimport malloc, free
cdef extern from "GpuIds.hpp":
    ctypedef struct GpuIds:
        # char* m_strName;
        int*  m_piDeviceIds; 
        int m_iCount; 

#TODO: Change from inline to stop duplication. Discussed on slack "python-questions" 28-Mar-2017
cdef inline void free_c_gpuids(GpuIds* c_gpuids):
    # free(c_gpuids.m_strName)
    if c_gpuids.m_piDeviceIds != NULL:
        free(c_gpuids.m_piDeviceIds)
    free(c_gpuids)

#TODO: Change from inline to stop duplication. Discussed on slack "python-questions" 28-Mar-2017
cdef inline GpuIds* convert_to_c_gpuids(p_gpuids):
    cdef GpuIds* c_gpuids =<GpuIds *>malloc(sizeof(GpuIds))
    if not c_gpuids:
        MemoryError("Error allocating memory for GPU IDs")
    if p_gpuids is not None:
        c_gpuids.m_iCount = len(p_gpuids)
    else:
        c_gpuids.m_iCount = 0
    
    if c_gpuids.m_iCount > 0:
        c_gpuids.m_piDeviceIds = <int*>malloc(c_gpuids.m_iCount * sizeof(int))
        if not c_gpuids.m_piDeviceIds:
            raise MemoryError("Error allocating memory for GPU IDs")
        for iI in range(c_gpuids.m_iCount):
            c_gpuids.m_piDeviceIds[iI] = p_gpuids.devices[iI]
    else:
        c_gpuids.m_iCount = 0
        c_gpuids.m_piDeviceIds = NULL

    return c_gpuids
